<?php

require_once dirname(__DIR__, 6) . '/VtbPay/vendor/autoload.php';

use \Vtbpay\Classes\Api\VtbApi,
    \Vtbpay\Classes\Common\VtbPayLogger,
    \Vtbpay\Classes\Exception\VtbPayException,
    \UmiCms\Service;

/**
 * A class intended for integration with the VTB payment system.
 * This handler allows you to set up and process payments via VTB in your online store.
 *
 * v 1.3.2
 *
 */
class VtbPayPayment extends payment
{
    /** @const string[] CURRENCIES поддерживаемые валюты */
    const CURRENCIES = ['RUB', 'RUR'];

    private $logger;


    /**
     * Checks whether the currency of the order is supported by the payment system
     *
     * @return bool Returns true if the order currency is supported, and false otherwise
     * @throws coreException Throws coreException when it encounters a system-level problem
     * @throws privateException Throws privateException if there's an issue with the private data
     */
    private function isSupportedCurrency() : bool
    {
        $currency = $this->getCurrencyCode();
        return in_array(strtoupper($currency), self::CURRENCIES);
    }


    /**
     * Checks if the order is suitable for processing by the payment system
     *
     * @param order $order The order to check
     * @return bool Returns true if the order is suitable, and false otherwise
     */
    public function isSuitable(order $order) : bool
    {
        return (
            $this->isSupportedCurrency() &&
            parent::isSuitable($order) &&
            ($order->getActualPrice() >= 0)
        );
    }


    /**
     * Retrieves the customer email from a customer object
     *
     * @param customer $customer The customer from which to retrieve the email
     * @return string The email of the customer
     * @throws publicException Throws publicException if the email is invalid
     */
    protected function getCustomerEmail(customer $customer)
    {
        $email = $customer->getEmail();

        if (
            !is_string($email) ||
            empty($email) ||
            !umiMail::checkEmail($email)
        ) {
            if (empty($email = $this->object->vtbpay_email_fiscal ?? nyll)) {
                throw new publicException(getLabel('error-payment-wrong-customer-email'));
            }
        }
        return $email;
    }


    /**
     * Retrieves the customer object based on the customer ID in the order
     *
     * @return customer Returns the customer object
     * @throws publicException Throws publicException if the customer ID is invalid
     */
    protected function getCustomer()
    {
        $customerId = $this->order->getCustomerId();
        $customerSource = selector::get('object')->id($customerId);

        if (!$customerSource instanceof iUmiObject) {
            throw new publicException(getLabel('error-payment-wrong-customer-id'));
        }

        return new customer($customerSource);
    }


    /**
     * Processes the order by creating a payment link and redirects to it
     *
     * @param string|null $template The template to use for processing
     * @throws Exception|publicException|privateException Throws exceptions when encountering errors during processing
     * @return null
     */
    public function process($template = null)
    {
        $this->setVtbpayLogger();
        $this->logger->setOption('additionalCommonText', 'payment-' . rand(1111, 9999));

        try {
            $this->order->order();
            $this->order->setPaymentStatus('initialized', true);

            $module = cmsController::getInstance()->getModule('emarket');
            if ($module) {
                $module->redirect($this->getPayUrl());
            }

        } catch (\Exception | publicException | privateException | VtbPayException $e) {
            // Handle exception and log error
            $context = [
                'file_exception' => $e->getFile(),
                'line_exception' => $e->getLine(),
            ];
            if (method_exists($e, 'getContext')) $context = array_merge($e->getContext(), $context);

            // Log the caught exception for debugging purposes.
            $this->logger->error(sprintf(
                __FUNCTION__ . ' > VtbPay Exception: %s; Order id: %s;',
                $e->getMessage(),
                $this->order->getId() ?: ''
            ), $context);

            throw $e;
        }
    }


    /**
     * Получает URL-адрес платежа для перенаправления.
     *
     * @return string URL-адрес платежа.
     */
    private function getPayUrl(): string
    {
        $orderId = $this->order->getId();
        // Order cost
        $amount = $this->formatPrice($this->order->getActualPrice());

        // Customer email
        $customer = $this->getCustomer();
        $customerEmail = $this->getCustomerEmail($customer);

        // Return Url
        $returnUrl = 'http://' . cmsController::getInstance()->getCurrentDomain()->getHost() .
                     '/emarket/gateway/' . $orderId . '/?request_type=return';

        // Since in this CMS the order identifier does not change when the order composition changes,
        // therefore, to create a new identifier, we associate the identifier with the price.
        $orderIdAndAmount = $orderId . '-' . $amount;

        $twoStage = (bool) ($this->object->vtbpay_two_stage ?? false);

        $enableFiscal = (bool) ($this->object->vtbpay_enable_fiscal ?? false);
        $items = [];
        if ($enableFiscal && !$twoStage) $items = $this->getItems();

        // Логирование входных данных
        $this->logger->debug(
            __FUNCTION__ . ' > getOrderLink - INPUT:', [
            'arguments' => [
                'orderIdAndAmount' => $orderIdAndAmount,
                'customerEmail' => $customerEmail,
                'time' => time(),
                'amount' => $amount,
                'returnUrl' => $returnUrl,
                'twoStage' => $twoStage,
                'items' => $items
            ]
        ]);

        $response = $this->getVtbApi()->getOrderLink(
            $orderIdAndAmount,
            $customerEmail,
            time(),
            $amount,
            $returnUrl,
            $twoStage,
            $items
        );

        // Логирование результата
        $this->logger->debug(
            __FUNCTION__ . ' > getOrderLink - OUTPUT:', [
            'response' => $response
        ]);

        return $response['object']['payUrl'] ?? '';
    }


    /**
     * Получает массив товаров заказа.
     *
     * @return array Массив товаров.
     */
    private function getItems(): array
    {
        $items = [];
        foreach ($this->order->getItems() as $key => $item) {
            $items[] = [
                'positionId' => ($key + 1),
                'name' => $item->getName(),
                'code' => (string) $item->getId(),
                'price' => floatval($item->getActualPrice()),
                'measure' => (int) $this->getValueSelectTypeObject('vtbpay-measure-fiscal') ?: 0,
                'quantity' => (int) $item->getAmount(),
                'taxParams' => [
                    'taxType' => $this->getValueSelectTypeObject('vtbpay-tax-type-fiscal') ?: 'none'
                ],
                'paymentType' => $this->getValueSelectTypeObject('vtbpay-payment-type-fiscal') ?: 'full_prepayment',
                'paymentSubject' => (int) $this->getValueSelectTypeObject('vtbpay-payment-subject-fiscal') ?: 1,
                'amount' => floatval($item->getActualPrice() * $item->getAmount())
            ];
        }

        $deliveryPrice = floatval($this->order->getValue('delivery_price'));
        if ($deliveryPrice > 0) {
            $delivery = selector::get('object')->id(
                $this->order->getValue('delivery_id')
            );

            if (!$delivery instanceof iUmiObject) {
                throw new expectObjectException(getLabel('error-unexpected-exception'));
            }
            $items[] = [
                'positionId' => ($key + 2),
                'name' => $delivery->getName() ?: 'Доставка',
                'price' => $deliveryPrice,
                'measure' => 0,
                'quantity' => 1,
                'taxParams' => [
                    'taxType' => $this->getValueSelectTypeObject('vtbpay-tax-type-fiscal') ?: 'none'
                ],
                'paymentType' => $this->getValueSelectTypeObject('vtbpay-payment-type-delivery-fiscal') ?: 'full_prepayment',
                'paymentSubject' => 4,
                'amount' => $deliveryPrice
            ];
        }

        return $items;
    }


    /**
     * Возвращает идентификатор заказа из запроса платежной системы.
     *
     * @return int
     */
    public static function getOrderId()
    {
        $mode = getRequest('request_type');
        if (!in_array($mode, ['return', 'webhook'])) throw new publicException('Undefined request type.');

        // Получение данных запроса в зависимости от типа
        $requestData = ($mode === 'return') ? $_REQUEST : json_decode(Service::Request()->getRawBody(), true);

        // Извлечение orderId
        $orderIdAndAmount = $requestData['orderId'] ?? $requestData['object']['orderId'] ?? '';
        if (empty($orderIdAndAmount)) {
            throw new publicException('Order ID not found.');
        }

        return (int) explode('-', $orderIdAndAmount)[0];
    }


    /**
     * Checks the status of the payment and updates the payment status accordingly
     *
     * @throws Exception|publicException|privateException Throws exceptions when encountering errors during polling
     */
    public function poll()
    {
        $this->setVtbpayLogger();

        $mode = getRequest('request_type');
        $this->logger->setOption('additionalCommonText', $mode . '-' . rand(1111, 9999));

        $orderId = $this->order->getId();
        // Order cost
        $amount = $this->formatPrice($this->order->getActualPrice());
        // Так как в данной CMS идентификатор заказа не меняется при изменении состава заказа,
        // поэтому для создания нового идентификатора мы связываем идентификатор с ценой.
        $orderIdAndAmount = $orderId . '-' . $amount;

        try {
            // Получение статуса оплаты
            $paymentStatus = $this->getPaymentStatus($orderIdAndAmount);

            // Проверка и смена статуса заказа
            $this->changePaymentStatus($orderIdAndAmount, $paymentStatus);

            $buffer = Service::Response()
                      ->getCurrentBuffer();
            $buffer->clear();
            if ($mode === 'return') {
                if (in_array($paymentStatus, ['PAID', 'PENDING'])) {
                    $host = $this->getSuccessUrl();
                }
                else {
                    $host = $this->getFailUrl();
                }
                $buffer->status('301 Moved Permanently');
                $buffer->redirect($host);
            }
            else {
                $buffer->contentType('application/json');
          			$buffer->push(json_encode(['status' => 'OK']));
            }
      			$buffer->end();

        } catch (\Exception | publicException | privateException | VtbPayException $e) {
            // Handle exception and log error
            $context = [
                'file_exception' => $e->getFile(),
                'line_exception' => $e->getLine(),
            ];
            if (method_exists($e, 'getContext')) $context = array_merge($e->getContext(), $context);

            // Log the caught exception for debugging purposes.
            $this->logger->error(sprintf(
                __FUNCTION__ . ' > VtbPay Exception: %s; Order ID and amount: %s;',
                $e->getMessage(),
                $orderIdAndAmount ?: ''
            ), $context);

            throw $e;
        }
    }


    /**
     * Получение статуса заказа WooCommerce из VTB API.
     *
     * @param string $orderIdAndAmount Идентификатор заказа + цена
     * @return string Значение статуса заказа, если оно существует, или пустая строка в противном случае.
     */
    private function getPaymentStatus(string $orderIdAndAmount): string
    {
        // Логирование входных данных
        $this->logger->debug(
            __FUNCTION__ . ' > getOrderInfo - INPUT:', [
            'arguments' => [
                'orderIdAndAmount' => $orderIdAndAmount,
            ]
        ]);

        $response = $this->getVtbApi()->getOrderInfo($orderIdAndAmount);

        // Логирование результата
        $this->logger->debug(
            __FUNCTION__ . ' > getOrderInfo - OUTPUT:', [
            'response' => $response
        ]);

        $this->order->setPaymentDocumentNumber(
            $response['object']['orderCode'] ?? ''
        );

        return $response['object']['status']['value'] ?? '';
    }


    /**
     * Смена статуса заказа в зависимости от статуса оплаты.
     *
     * @param string $paymentStatus Статус оплаты
     * @return void
     * @throws LocalizedException
     */
    private function changePaymentStatus(string $orderIdAndAmount, string $paymentStatus)
    {
        $availableStatuses = [
           'PAID' => 'accepted',
           'PENDING' => 'waiting_confirmation',
           'REFUNDED' => 'refund'
        ];

        $newPaymentStatus = $availableStatuses[$paymentStatus] ?? 'declined';

        $this->logger->debug(
            __FUNCTION__ . ' - change_status:', [
            'newPaymentStatus' => $newPaymentStatus
        ]);

        // не сравниваем старый статус с новым, так это уже есть в setPaymentStatus
        $this->order->setPaymentStatus($newPaymentStatus);
        // Иначе корзина не очищается при холдировании
        if ($newPaymentStatus === 'waiting_confirmation') {
            $this->order->setValue('order_date', time());
            $this->order->setOrderStatus('waiting');
        }
        $this->order->commit();
    }


    /**
     * Получает значение объекта типа выпадающий список
     *
     * @param string $typeName Название типа
     *
     * @return string Значение
     */
    private function getValueSelectTypeObject($typeName): string
    {
        $propertyName = str_replace('-', '_', $typeName);
        $object = umiObjectsCollection::getInstance()->getObject($this->object->$propertyName ?? NULL);
        if (empty($object)) return '';
        return str_replace($typeName . '-', '', $object->getGUID());
    }


    /**
     * Получает объект для работы с API ВТБ.
     *
     * @return VtbApi Объект для работы с API ВТБ.
     */
    public function getVtbApi(): VtbApi
    {
        return new VtbApi(
            $this->object->vtbpay_client_id,
            $this->object->vtbpay_client_secret,
            (bool) $this->object->vtbpay_test_mode,
            $this->object->vtbpay_merchant_authorization ?: ''
        );
    }


    /**
     * Устанавливает объект логирования для модуля.
     */
    public function setVtbpayLogger(): void
    {
        $logging = (bool) $this->object->vtbpay_logging;

        $sysLogPath = mainConfiguration::getInstance()->includeParam('sys-log-path');
        $logPath = sprintf('%s/payment/%s/', rtrim($sysLogPath, '/'), str_replace('Payment', '', get_class($this)));
        $fileName = 'vtbpay-' . date('d-m-Y');

        $logger = new \umiLogger($logPath);
        $logger->setFileName($fileName);
        $logger->separateByIp(false);

        $this->logger = VtbPayLogger::getInstance()
                        ->setLogFilePath($logPath . $fileName . '.log')
                        ->setCustomRecording(function($message) use ($logger)
                        {
                            // Логирование ошибки
                            $logger->push($message . PHP_EOL, false);
                            $logger->save();

                        }, VtbPayLogger::LOG_LEVEL_ERROR)
                        ->setCustomRecording(function($message) use ($logger, $logging)
                        {
                            // Логирование сообщения дебага
                            if ($logging) {
                                $logger->push($message . PHP_EOL, false);
                                $logger->save();
                            }

                        }, VtbPayLogger::LOG_LEVEL_DEBUG);
    }
}
